#define clearPutty()    printf( "\x1b[2J")
#define homePutty()     printf( "\x1b[H")
#define PB_CLK          sys_clock
#include "config.h"

#include <stdlib.h>
#include <math.h>
#include "plib.h"
#include <xc.h>
#include "sd_routines.h"
#include "fat32.h"

void setupUART(void) {
    PPSInput (2, U2RX, RPA1); //Assign U2RX to pin RPA1 -- Physical pin 3 on 28 PDIPS
    PPSOutput(4, RPB10, U2TX); //Assign U2TX to pin RPB10 -- Physical pin 21 on 28 PDIP

    // white on adafruit 954 is RX in - so connect to pin 21
    // green on adafruit 954 is TX out - so connect to pin 3

    UARTConfigure(UART2, UART_ENABLE_PINS_TX_RX_ONLY);
    UARTSetLineControl(UART2, UART_DATA_SIZE_8_BITS | UART_PARITY_NONE | UART_STOP_BITS_1);
    UARTSetDataRate(UART2, PB_CLK, BAUDRATE);
    UARTEnable(UART2, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX | UART_TX));
}

//**************************************************
//Function to receive a single byte
//*************************************************
unsigned char receiveByte( void )
{
	unsigned char data;
	
	while(!DataRdyUART2()); 	// Wait for incomming data
	
	data = ReadUART2();
	
	return(data);
}

//***************************************************
//Function to transmit a single byte
//***************************************************
void transmitByte( unsigned char data )
{
	do{while(!U2STAbits.TRMT); WriteUART2(data);}while(0);
}

//***************************************************
//Function to transmit a string in RAM
//***************************************************
void transmitString(unsigned char* string)
{
  while (*string)
   transmitByte(*string++);
}

//***************************************************
void transmitHex( unsigned char dataType, unsigned long data )
{
unsigned char count, i, temp;
unsigned char dataString[] = "0x        ";

if (dataType == 0) count = 2;
if (dataType == 1) count = 4;
if (dataType == 2) count = 8;

for(i=count; i>0; i--)
{
  temp = data % 16;
  if((temp>=0) && (temp<10)) dataString [i+1] = temp + 0x30;
  else dataString [i+1] = (temp - 10) + 0x41;

  data = data/16;
}

transmitString (dataString);
}

//***************************************************
//Function to transmit a string in Flash
//***************************************************
/*  SDO:    RB13    (TRIS defined in HardwareProfile.h and PPS defined in SD-SPI.c)
 *          physical pin 24
 *  SDI:    RB11    (TRIS defined in HardwareProfile.h and PPS defined in SD-SPI.c)
 *          physical pin 22
 *  SCK:    RB14    (TRIS defined in HardwareProfie.h)
 *          physical pin 25
 *  CS:     RB2     (defined in HardwareProfile.h)*
 * 
 *  #define TRIS_SCK TRISBbits.TRISB14
 *  #define TRIS_SDI TRISBbits.TRISB11
 *  #define TRIS_SDO TRISBbits.TRISB13
 *  #define TRIS_CS  TRISBbits.TRISB2   */

void setupSPI() {
    volatile SpiChannel spiChn = SPI_CHANNEL1; // the SPI channel to use
    volatile int spiClkDiv = 160; // 250k Hz speed for SD initialization 
    PPSOutput(3, RPB13, SDO1);    // RB13 as SDO1
    PPSInput(2, SDI1, RPB11);     // RB11 as SDI1
    
    mPORTBSetPinsDigitalOut(BIT_2); //RB2 as CS
    mPORTBSetBits(BIT_2);           //High for deselect
    SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE8 | SPI_OPEN_MSTEN , spiClkDiv | SPI_OPEN_SMP_END);
}

unsigned char SPI_transmit(unsigned char data)
{
// Start transmission
  SPI1BUF = data;

// Wait for transmission complete
  while(!SPI1STATbits.SPITBE);
  data = SPI1BUF ;
  return data;
}

unsigned char SPI_receive(void)
{
  unsigned char data;
// Wait for reception complete
  SPI1BUF = 0xff;

  while(!SPI1STATbits.SPIRBF);
  data = SPI1BUF;
  return data;
}

//***
//******************************************************************
unsigned char SD_init(void)
{
unsigned char i, response, SD_version;
unsigned int retry=0 ;

 for(i=0;i<10;i++)
      SPI_transmit(0xff);   //80 clock pulses spent before sending the first command

SD_CS_ASSERT;
//printf("sd_cs assert\n\r");
do
{
   
   response = SD_sendCommand(GO_IDLE_STATE, 0); //send 'reset & go idle' command
   retry++;
   if(retry>0x20){ 
      printf("time out, card not detected\n\r");
   	  return 1;   //time out, card not detected
   }
   
} while(response != 0x01);

SD_CS_DEASSERT;
//transmitString("sd deassert\n");

SPI_transmit (0xff);
SPI_transmit (0xff);

retry = 0;

SD_version = 2; //default set to SD compliance with ver2.x; 
				//this may change after checking the next command
do
{
response = SD_sendCommand(SEND_IF_COND,0x000001AA); //Check power supply status, mendatory for SDHC card
retry++;
if(retry>0xfe) 
   {
	  //TX_NEWLINE;
	  SD_version = 1;
	  cardType = 1;
	  break;
   } //time out

}while(response != 0x01);

retry = 0;

do
{
response = SD_sendCommand(APP_CMD,0); //CMD55, must be sent before sending any ACMD command
response = SD_sendCommand(SD_SEND_OP_COND,0x40000000); //ACMD41

retry++;
if(retry>0xfe) 
   {
      
	  return 2;  //time out, card initialization failed
   } 

}while(response != 0x00);


retry = 0;
SDHC_flag = 0;

if (SD_version == 2)
{ 
   do
   {
	 response = SD_sendCommand(READ_OCR,0);
	 retry++;
	 if(retry>0xfe) 
     {
       //TX_NEWLINE;
	   cardType = 0;
	   break;
     } //time out

   }while(response != 0x00);

   if(SDHC_flag == 1) cardType = 2;
   else cardType = 3;
}

printf("sd version = %d, cardType = %d\n\r",SD_version,cardType);
printf("\n");
//; //disable CRC; deafault - CRC disabled in SPI mode
//SD_sendCommand(SET_BLOCK_LEN, 512); //set block size to 512; default size is 512


return 0; //successful return
}


//******************************************************************
unsigned char SD_sendCommand(unsigned char cmd, unsigned long arg)
{
unsigned char response, retry=0, status;

//SD card accepts byte address while SDHC accepts block address in multiples of 512
//so, if it's SD card we need to convert block address into corresponding byte address by 
//multipying it with 512. which is equivalent to shifting it left 9 times
//following 'if' loop does that

if(SDHC_flag == 0)		
{if(cmd == READ_SINGLE_BLOCK     ||
   cmd == READ_MULTIPLE_BLOCKS  ||
   cmd == WRITE_SINGLE_BLOCK    ||
   cmd == WRITE_MULTIPLE_BLOCKS ||
   cmd == ERASE_BLOCK_START_ADDR|| 
   cmd == ERASE_BLOCK_END_ADDR ) 
   {
     arg = arg << 9;
   }	   
}
SD_CS_ASSERT;

SPI_transmit(cmd | 0x40); //send command, first two bits always '01'
SPI_transmit(arg>>24);
SPI_transmit(arg>>16);
SPI_transmit(arg>>8);
SPI_transmit(arg);

if(cmd == SEND_IF_COND)	 //it is compulsory to send correct CRC for CMD8 (CRC=0x87) & CMD0 (CRC=0x95)
  SPI_transmit(0x87);    //for remaining commands, CRC is ignored in SPI mode
else 
  SPI_transmit(0x95); 

while((response = SPI_receive())== 0xff){ //wait response
    
    if(retry++ > 0xfe) break; //time out error
}
//printf("cmd = %d , response = %d  in  SD_sendCommand\n\r",cmd,response);

if(response == 0x00 && cmd == 58)  //checking response of CMD58
{
  status = SPI_receive() & 0x40;     //first byte of the OCR register (bit 31:24)
  if(status == 0x40) SDHC_flag = 1;  //we need it to verify SDHC card
  else SDHC_flag = 0;

  SPI_receive(); //remaining 3 bytes of the OCR register are ignored here
  SPI_receive(); //one can use these bytes to check power supply limits of SD
  SPI_receive(); 
}


SPI_receive(); //extra 8 CLK
SD_CS_DEASSERT;

return response; //return state
}


unsigned char SD_erase (unsigned long startBlock, unsigned long totalBlocks)
{
unsigned char response;

response = SD_sendCommand(ERASE_BLOCK_START_ADDR, startBlock); //send starting block address
if(response != 0x00) //check for SD status: 0x00 - OK (No flags set)
  return response;

response = SD_sendCommand(ERASE_BLOCK_END_ADDR,(startBlock + totalBlocks - 1)); //send end block address
if(response != 0x00)
  return response;

response = SD_sendCommand(ERASE_SELECTED_BLOCKS, 0); //erase all selected blocks
if(response != 0x00)
  return response;

return 0; //normal return
}


unsigned char SD_readSingleBlock(unsigned long startBlock)
{
unsigned char response;
unsigned int i, retry=0;

 response = SD_sendCommand(READ_SINGLE_BLOCK, startBlock); //read a Block command
 
 if(response != 0x00) 
 {
     printf("no response in SD_readSingleBlock\n\r");
     return response; //check for SD status: 0x00 - OK (No flags set)
 }

SD_CS_ASSERT;

retry = 0;
while(SPI_receive() != 0xfe) //wait for start block token 0xfe (0x11111110)
  if(retry++ > 0xfffe){SD_CS_DEASSERT; return 1;} //return if time-out

for(i=0; i<512; i++) //read 512 bytes
  buffer[i] = SPI_receive();

SPI_receive(); //receive incoming CRC (16-bit), CRC is ignored here
SPI_receive();

SPI_receive(); //extra 8 clock pulses
SD_CS_DEASSERT;

return 0;
}


//******************************************************************
unsigned char SD_writeSingleBlock(unsigned long startBlock)
{
unsigned char response;
unsigned int i, retry=0;

 response = SD_sendCommand(WRITE_SINGLE_BLOCK, startBlock); //write a Block command
  
 if(response != 0x00) 
 {
  printf("no response in SD_writeSingleBlock\n\r");
  return response; //check for SD status: 0x00 - OK (No flags set)
 }
SD_CS_ASSERT;  // CS set to low 

//SPI_transmit(0xff);
//SPI_transmit(0xff);
SPI_transmit(0xfe);     //Send start block token 0xfe (0x11111110)

for(i=0; i<512; i++)    //send 512 bytes data
  SPI_transmit(buffer[i]);

SPI_transmit(0xff);     //transmit dummy CRC (16-bit), CRC is ignored here
SPI_transmit(0xff);

response = SPI_receive();
response = SPI_receive();

if( (response & 0x1f) != 0x05) //response= 0xXXX0AAA1 ; AAA='010' - data accepted
{                              //AAA='101'-data rejected due to CRC error
  SD_CS_DEASSERT;              //AAA='110'-data rejected due to write error
 // printf("a xi ba");
  return response;
}

while(!SPI_receive()) //wait for SD card to complete writing and get idle
if(retry++ > 0xfffe){SD_CS_DEASSERT; return 1;}

SD_CS_DEASSERT;
SPI_transmit(0xff);   //just spend 8 clock cycle delay before reasserting the CS line
SD_CS_ASSERT;         //re-asserting the CS line to verify if card is still busy

while(!SPI_receive()) //wait for SD card to complete writing and get idle
   if(retry++ > 0xfffe){SD_CS_DEASSERT; return 1;}
SD_CS_DEASSERT;

return 0;
}
